{ File   : Manager.pas
  Author : Deniz Oezmen
  Created: 2007-07-17
  Changed: 2007-08-04

  IMA ADPCM compression/decompression framework.
}
unit Manager;

interface

uses
  Shared, SysUtils;

const
  IMA_ADPCM_Format_Invalid: TGUID = '{00000000-0000-0000-0000-000000000000}';

type
  TIMA_ADPCM_Manager = class(TObject)
  protected
    FFormatType  : TGUID;
    FFunctions   : TPluginFunctions;
    FBlockSize   : Longint;
    FNumChannels : Longint;
    FSampPerBlock: Longint;
    FStates      : TIMA_ADPCM_States;
    FSampBuffer  : TIMA_ADPCM_SampleBuffer;
    FSampBufIndex: Longint;
    FCodeBuffer  : TIMA_ADPCM_CodeBuffer;
    FCodeBufIndex: Longint;
    FInitSuccess : Boolean;
  public
    constructor Create(BlockSize, SampPerBlock, NumChannels: Longint;
      const FormatType: TGUID);
    destructor Destroy; override;
    procedure AssignSampleBuffer(const SampleBuffer: TIMA_ADPCM_SampleBuffer);
    procedure UnassignSampleBuffer;
    procedure AssignCodeBuffer(const CodeBuffer: TIMA_ADPCM_CodeBuffer);
    procedure UnassignCodeBuffer;
    procedure GetSampleBuffer(out SampleBuffer: TIMA_ADPCM_SampleBuffer);
    procedure GetCodeBuffer(out CodeBuffer: TIMA_ADPCM_CodeBuffer);
    function GetCompressedBlockSize: Longint;
    function GetDecompressedBlockSize: Longint;
    function GetCompressedSize: Longint; overload;
    function GetCompressedSize(NumSamples: Longint): Longint; overload;
    function GetDecompressedSize: Longint; overload;
    function GetDecompressedSize(CodeSize: Longint): Longint; overload;
    function CompressBlock: Longint;
    function DecompressBlock: Longint;
    procedure Compress;
    procedure Decompress;
  end;

implementation

constructor TIMA_ADPCM_Manager.Create(BlockSize, SampPerBlock,
  NumChannels: Longint; const FormatType: TGUID);
var
  i: Longint;
begin
  inherited Create;

  FFormatType := FormatType;

  // get plugin functions
  GetPluginFunctions(FFormatType, FFunctions);
  with FFunctions do
    FInitSuccess := Assigned(CorrectParams) and Assigned(GetCompressedSize) and
      Assigned(GetDecompressedSize) and Assigned(CompressToBuffer) and
      Assigned(DecompressFromBuffer);
  if not FInitSuccess then
    FFormatType := IMA_ADPCM_Format_Invalid;

  FBlockSize := BlockSize;
  FNumChannels := NumChannels;
  FSampPerBlock := SampPerBlock;

  if Assigned (FFunctions.CorrectParams) then
    FFunctions.CorrectParams(FBlockSize, FSampPerBlock, FNumChannels);

  SetLength(FStates, FNumChannels);
  for i := 0 to High(FStates) do
    with FStates[i] do
    begin
      Index := 0;
      PrevSample := 0
    end;

  FSampBuffer := nil;
  FSampBufIndex := 0;

  FCodeBuffer := nil;
  FCodeBufIndex := 0
end;

destructor TIMA_ADPCM_Manager.Destroy;
var
  i: Longint;
begin
  FFormatType := IMA_ADPCM_Format_Invalid;

  FBlockSize := 0;
  FNumChannels := 0;
  FSampPerBlock := 0;

  for i := 0 to High(FStates) do                        
    with FStates[i] do
    begin
      Index := 0;
      PrevSample := 0
    end;
  Finalize(FStates);
  SetLength(FStates, 0);
  FStates := nil;

  Finalize(FSampBuffer);
  SetLength(FSampBuffer, 0);
  FSampBuffer := nil;
  FSampBufIndex := 0;

  Finalize(FCodeBuffer);
  FCodeBuffer := nil;
  FCodeBufIndex := 0;

  FInitSuccess := False;

  inherited Destroy
end;

procedure TIMA_ADPCM_Manager.AssignSampleBuffer(
  const SampleBuffer: TIMA_ADPCM_SampleBuffer);
begin
  FSampBuffer := SampleBuffer;
  FSampBufIndex := 0
end;

procedure TIMA_ADPCM_Manager.UnassignSampleBuffer;
begin
  FSampBuffer := nil;
  FSampBufIndex := 0
end;

procedure TIMA_ADPCM_Manager.AssignCodeBuffer(
  const CodeBuffer: TIMA_ADPCM_CodeBuffer);
begin
  FCodeBuffer := CodeBuffer;
  FCodeBufIndex := 0
end;

procedure TIMA_ADPCM_Manager.UnassignCodeBuffer;
begin
  FCodeBuffer := nil;
  FCodeBufIndex := 0
end;

procedure TIMA_ADPCM_Manager.GetSampleBuffer(
  out SampleBuffer: TIMA_ADPCM_SampleBuffer);
begin
  SampleBuffer := FSampBuffer
end;

procedure TIMA_ADPCM_Manager.GetCodeBuffer(
  out CodeBuffer: TIMA_ADPCM_CodeBuffer);
begin
  CodeBuffer := FCodeBuffer
end;

function TIMA_ADPCM_Manager.GetCompressedBlockSize: Longint;
begin
  if Assigned (FFunctions.CorrectParams) then
    FFunctions.CorrectParams(FBlockSize, FSampPerBlock, FNumChannels);

  Result := FBlockSize
end;

function TIMA_ADPCM_Manager.GetDecompressedBlockSize: Longint;
begin
  if Assigned (FFunctions.CorrectParams) then
    FFunctions.CorrectParams(FBlockSize, FSampPerBlock, FNumChannels);

  Result := FSampPerBlock
end;

function TIMA_ADPCM_Manager.GetCompressedSize: Longint;
begin
  if Assigned (FFunctions.CorrectParams) then
    FFunctions.CorrectParams(FBlockSize, FSampPerBlock, FNumChannels);

  if FInitSuccess then
    Result := FFunctions.GetCompressedSize(FBlockSize, FSampPerBlock,
      FNumChannels, Length(FSampBuffer))
  else
    Result := 0
end;

function TIMA_ADPCM_Manager.GetCompressedSize(NumSamples: Longint): Longint;
begin
  if Assigned (FFunctions.CorrectParams) then
    FFunctions.CorrectParams(FBlockSize, FSampPerBlock, FNumChannels);

  if FInitSuccess then
    Result := FFunctions.GetCompressedSize(FBlockSize, FSampPerBlock,
      FNumChannels, NumSamples)
  else
    Result := 0
end;

function TIMA_ADPCM_Manager.GetDecompressedSize: Longint;
begin
  if Assigned (FFunctions.CorrectParams) then
    FFunctions.CorrectParams(FBlockSize, FSampPerBlock, FNumChannels);

  if FInitSuccess then
    Result := FFunctions.GetDecompressedSize(FBlockSize, FSampPerBlock,
      FNumChannels, Length(FCodeBuffer))
  else
    Result := 0
end;

function TIMA_ADPCM_Manager.GetDecompressedSize(CodeSize: Longint): Longint;
begin
  if Assigned (FFunctions.CorrectParams) then
    FFunctions.CorrectParams(FBlockSize, FSampPerBlock, FNumChannels);

  if FInitSuccess then
    Result := FFunctions.GetDecompressedSize(FBlockSize, FSampPerBlock,
      FNumChannels, CodeSize)
  else
    Result := 0
end;

function TIMA_ADPCM_Manager.CompressBlock: Longint;
begin
  if Assigned (FFunctions.CorrectParams) then
    FFunctions.CorrectParams(FBlockSize, FSampPerBlock, FNumChannels);

  SetLength(FCodeBuffer, FBlockSize);
  FillChar(FCodeBuffer[0], Length(FCodeBuffer), 0);

  if FInitSuccess then
    SetLength(FCodeBuffer, FFunctions.CompressToBuffer(FSampBuffer,
      FSampBufIndex, FCodeBuffer, FCodeBufIndex, FStates, FBlockSize,
      FSampPerBlock, FNumChannels))
  else
    SetLength(FCodeBuffer, 0);

  Result := Length(FCodeBuffer)
end;

function TIMA_ADPCM_Manager.DecompressBlock: Longint;
begin
  if Assigned (FFunctions.CorrectParams) then
    FFunctions.CorrectParams(FBlockSize, FSampPerBlock, FNumChannels);

  SetLength(FSampBuffer, FSampPerBlock);
  FillChar(FSampBuffer[0], Length(FSampBuffer) * 2, 0);

  if FInitSuccess then
    SetLength(FSampBuffer, FFunctions.DecompressFromBuffer(FSampBuffer,
      FSampBufIndex, FCodeBuffer, FCodeBufIndex, FStates, FBlockSize,
      FSampPerBlock, FNumChannels))
  else
    SetLength(FSampBuffer, 0);

  Result := Length(FSampBuffer)
end;

procedure TIMA_ADPCM_Manager.Compress;
begin
  if not FInitSuccess then
    Exit;

  if Assigned (FFunctions.CorrectParams) then
    FFunctions.CorrectParams(FBlockSize, FSampPerBlock, FNumChannels);

  SetLength(FCodeBuffer, GetCompressedSize);
  FillChar(FCodeBuffer[0], Length(FCodeBuffer), 0);

  FCodeBufIndex := 0;
  FSampBufIndex := 0;

    while (FFunctions.CompressToBuffer(FSampBuffer, FSampBufIndex, FCodeBuffer,
      FCodeBufIndex, FStates, FBlockSize, FSampPerBlock, FNumChannels) > 0) do
      ;
end;

procedure TIMA_ADPCM_Manager.Decompress;
begin
  if not FInitSuccess then
    Exit;

  if Assigned (FFunctions.CorrectParams) then
    FFunctions.CorrectParams(FBlockSize, FSampPerBlock, FNumChannels);

  SetLength(FSampBuffer, GetDecompressedSize);
  FillChar(FSampBuffer[0], Length(FSampBuffer) * 2, 0);

  FCodeBufIndex := 0;
  FSampBufIndex := 0;

  while (FFunctions.DecompressFromBuffer(FSampBuffer, FSampBufIndex,
    FCodeBuffer, FCodeBufIndex, FStates, FBlockSize, FSampPerBlock,
    FNumChannels) > 0) do
    ;
end;

end.

